#!/usr/bin/env python
"""This module has tests for the bitunflip functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import unittest
from pathlib import Path
from unittest.mock import call, patch

from hiproc import bitunflip as bf

# from .utils import resource_check as rc

# # Hardcoding this, but I sure would like a better solution.
# HiRISE_img = Path('test-resources') / 'PSP_010502_2090_RED5_0.img'
# img = HiRISE_img
#
#
# class TestResources(unittest.TestCase):
#     '''Establishes that the test image exists.'''
#
#     def test_resources(self):
#         (truth, test) = rc(img)
#         self.assertEqual(truth, test)


class TestBasic(unittest.TestCase):
    def test_get_range(self):
        start = 2
        stop = 5
        truth = range(start, stop, 1)
        self.assertEqual(truth, bf.get_range(start, stop))

        start = 2.1
        stop = 4.8
        self.assertEqual(truth, bf.get_range(start, stop))

        start = 9.1
        truth = range(10, 4, -1)
        self.assertEqual(truth, bf.get_range(start, stop))


class TestHist(unittest.TestCase):
    def setUp(self):
        HistRow = collections.namedtuple("HistRow", ["DN", "Pixels"])
        self.hist = [
            HistRow(1, 1),  # 0 *
            HistRow(2, 2),  # 1 **
            HistRow(5, 3),  # 2 ***
            HistRow(6, 4),  # 3 ****
            HistRow(7, 3),  # 4 ***
            HistRow(8, 2),  # 5 **
            HistRow(10, 1),  # 6 *
            HistRow(11, 2),  # 7 **
            HistRow(12, 2),  # 8 **
            # Biggest gap here between 12 and 20
            HistRow(20, 1),
        ]  # 9 *

    def test_find_gap(self):
        self.assertEqual(13, bf.find_gap(self.hist, 1, 20))
        self.assertEqual(3, bf.find_gap(self.hist, 1, 20, findfirst=True))
        self.assertEqual(9, bf.find_gap(self.hist, 5, 12))

    def test_find_min_dn(self):
        self.assertEqual(1, bf.find_min_dn(self.hist, 1, 20))
        self.assertEqual(10, bf.find_min_dn(self.hist, 2, 20))
        self.assertEqual(20, bf.find_min_dn(self.hist, 20, 1))

    @patch("hiproc.bitflips.shutil.copyfile")
    @patch("hiproc.bitflips.isis.mask")
    @patch("hiproc.bitflips.isis.algebra")
    @patch("hiproc.bitflips.isis.handmos")
    def test_subtract_over_thresh(self, m_hand, m_alg, m_mask, m_copyfile):
        in_p = Path("in.dummy")
        out_p = Path("out.dummy")
        bf.subtract_over_thresh(in_p, out_p, thresh=1, delta=2, keep=True)

        m_copyfile.assert_called_once()

        m_mask.assert_called_once()
        t_path = in_p.with_suffix(".threshmask.cub")
        d_path = in_p.with_suffix(".delta.cub")
        mask_args = {"from": in_p, "min": 1, "to": t_path}
        self.assertEqual(m_mask.call_args_list, [call(**mask_args)])

        m_alg.assert_called_once()
        self.assertEqual(
            m_alg.call_args_list,
            [call(t_path, a=0, c=-2, from2=in_p, op="add", to=d_path)],
        )

        m_hand.assert_called_once()
        self.assertEqual(m_hand.call_args_list, [call(d_path, mosaic=out_p)])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import collections
import csv
import itertools

from pathlib import Path


class FlatFile(collections.abc.Sequence):
    """Reads the output from the flat file created by hijitreg and provides
    it as a sequence.

    The resulting FlatFile object primarily behaves like a list, where
    that list represents the rows of the ISIS hijitreg output.  Each of those
    rows is a :func:`collections.namedtuple` which contains the elements
    of each row, referenced by their column names.

    The FlatFile object also has some dictionary-like capabilities, in
    order to get at the values listed in the ISIS hijitreg output in the
    section before the numerical output.
    """

    def __init__(self, flatinfo):
        self.flatinfo = flatinfo

        try:  # Assume it is a string?
            (self.dictionary, self.headers, self.the_list) = self.parse(
                flatinfo
            )
        except ValueError:
            # Maybe it is a file path?
            (self.dictionary, self.headers, self.the_list) = self.parse(
                Path(flatinfo).read_text()
            )

    def __str__(self):
        return str(self.dictionary)

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.flatinfo}')"

    def __len__(self):
        return len(self.the_list)

    def __getitem__(self, key):
        try:
            return self.dictionary[key]
        except KeyError:
            return self.the_list[key]

    def __iter__(self):
        return self.the_list.__iter__()

    def __contains__(self, item):
        if item in self.dictionary:
            return True
        else:
            return item in self.the_list

    def keys(self):
        """Gets the keys from the initial portion of the flat output file.

           These will be items like 'FROM', 'MATCH', 'RegFile', etc.
        """
        return self.dictionary.keys()

    def values(self):
        """Gets the values from the initial portion of the flat output file."""
        return self.dictionary.values()

    @staticmethod
    def parse(flatinfo: str) -> tuple:
        """Takes a string (expecting the output of ISIS ``hijitreg``), and
        parses the output.

        A three-element namedtuple is returned: the first element
        is a *dictionary* of the name:value information at the
        top of the file, the second element is a *list* of the
        of the fields that decorate the top of the data
        rows, and the third element is a *list* of FlatRow
        :func:`collections.namedtuple` objects that represent each
        row of the flat output.

        The contents of the file that results from ISIS ``hijitreg``
        look like this::

            #          Hijitreg ISIS Application Results
            #    Coordinates are (Sample, Line) unless indicated
            #           RunDate:  2012-04-11T20:02:49
            #
            #    ****  Image Input Information ****
            #  FROM:  /HiRISE/Users/audrie/pipe_dev_sandbox/HiRISE/Data/ResolveJitter/ESP/ORB_016200_016299/ESP_016213_2315/ESP_016213_2315_RED4.prehijack.cub
            #    Lines:       70000
            #    Samples:     2048
            #    FPSamp0:     0
            #    SampOffset:  0
            #    LineOffset:  0
            #    CPMMNumber:  5
            #    Summing:     1
            [... other lines like this ...]


            #  MATCH: /HiRISE/Users/audrie/pipe_dev_sandbox/HiRISE/Data/ResolveJitter/ESP/ORB_016200_016299/ESP_016213_2315/ESP_016213_2315_BG12.prehijack.cub
            #    Lines:       70000
            #    Samples:     2048
            #    FPSamp0:     0
            #    SampOffset:  0
            [... other lines like this ...]

            #   Total Registers:  10145 of 10500
            #   Number Suspect:   0
            #   Average Sample Offset: 0.3586  StdDev: 0.3605
            #   Average Line Offset:   3.9526 StdDev: 1.5403

            #  Column Headers and Data
                        FromTime  FromSamp  FromLine           MatchTime MatchSamp MatchLine        RegSamp        RegLine   RegCorr      B0_Offset       B1_Slope   B_RCorr
              316426108.25476718       341       210  316426108.15863681       341       210       340.6611       204.2240  0.883177       0.024032       0.273207  0.894237
              316426108.25476718      1023       210  316426108.15863681      1023       210      1022.5484       204.4079  0.805408       0.020917       0.286734  0.839953
              316426108.25476718      1705       210  316426108.15863681      1705       210      1704.6887       204.6171  0.823995       0.026994       0.258996  0.853913
            [... more comma-separated lines like above ...]

        But this function sees it like this::

            #          Hijitreg ISIS Application Results
            #    Coordinates are (Sample, Line) unless indicated
            #           k:  v
            #
            #    ****  Image Input Information ****
            #  k: v_FROM
            #  [here the key in the parsed dict will be 'FROM']
            #  [the value of the 'FROM' key will be another dictionary]
            #  [that dictionary will contain a 'path' key with the filename]
            #  [the other key-value pairs will come from these items]
            #    Lines:       70000
            #    Samples:     2048
            #    FPSamp0:     0
            #    SampOffset:  0
            #    LineOffset:  0
            #    CPMMNumber:  5
            #    Summing:     1
            [... other lines like this ...]

            #  k: v_MATCH
            # [again the MATCH key will have a similar dictionary of values.

            [... other lines like this ...]

            #   k: v
            #   k: v
            #   k: v k_StdDev: v
            #   k: v k_StdDev: v

            #  Column Headers and Data
            h h h h h h h h h h h h
            n n n n n n n n n n n n
            n n n n n n n n n n n n
            n n n n n n n n n n n n
            [... more whitespace-separated lines like above ...]

        Where each of the letters above is a string value that the parser
        reads.

        First, it takes all of the ``k`` and ``v`` elements and saves them
        as the keys and values in the returned dictionary.
        For the special FROM and MATCH keys, their values will each be a
        dictionary with their printed elements and a special key named
        path which will contain the string after "FROM:" or "MATCH:".

        If there are "StdDev" elements on a line that follows a key, they
        will be saved with a key value of the previous key plus " StdDev"
        such that a line like this::

            Average Sample Offset: 0.3586  StdDev: 0.3605

        Will be turned into two entries in the dictionary, one with a key
        value of "Average Sample Offset" and one with
        "Average Sample Offset StdDev".

        Second, the ``h`` elements are returned as the list of
        fieldnames.

        Third, it reads the lines with ``n`` and stores them as
        ``namedtuples`` in the returned list.
        """
        d = dict()
        flat_rows = []

        # This would be so much easier if they just wrote out JSON or something.
        # Gong to have to get a little weird to deal with the FROM and MATCH
        # dicts.
        indent_level = 0
        prev_dict = None
        for line in filter(
            lambda x: ":" in x, str(flatinfo).splitlines()
        ):
            (k, colon, v) = line.lstrip("#").partition(":")
            leading_spaces = len(k) - len(k.lstrip(' '))
            the_key = k.strip()
            the_value = v.strip()
            if leading_spaces < indent_level:
                prev_dict = None

            # Does this have a weird trailing standard deviation?
            val, stdsep, std = the_value.partition("StdDev:")
            if stdsep != "":
                the_value = val.strip()

            if prev_dict is not None:
                d[prev_dict].setdefault(the_key, the_value)

            else:
                if the_key == "FROM" or the_key == "MATCH":
                    prev_dict = the_key
                    the_value = {"path": the_value}

                d.setdefault(the_key, the_value)

            # Since this had a standard deviation, add it as a key:
            if stdsep != "":
                d.setdefault(the_key + " " + stdsep.strip(":"), std.strip())

            indent_level = leading_spaces

        if len(d) == 0:
            raise ValueError(
                "Could not extract any header information from the flat file."
            )

        # Read the data rows.
        dialect = csv.Dialect
        dialect.delimiter = ' '
        dialect.skipinitialspace = True
        dialect.quoting = csv.QUOTE_NONE
        dialect.lineterminator = '\n'

        reader = csv.DictReader(itertools.filterfalse(lambda x:
                                                      x.startswith('#') or
                                                      x.isspace() or
                                                      len(x) == 0,
                                                      flatinfo.splitlines()),
                                dialect=dialect)

        for row in reader:
            flat_rows.append(row)

        FlatParsed = collections.namedtuple(
            "FlatParsed", ["info", "fieldnames", "data"]
        )
        return FlatParsed(d, reader.fieldnames, flat_rows)



#!/usr/bin/env python
"""Functions to work with HiRISE .img files.
"""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Reuse is permitted under the terms of the license.
# The LICENSE file is at the top level of this library.

import os
import statistics

import numpy as np
import pvl
import kalasiris as isis


class LUT_Table:
    """HiRISE Look-Up Table class.

    This is very specific to the LUT tables in HiRISE EDRs.
    """

    def __init__(self, lut):
        if len(lut) != 256 and len(lut) != 1:
            raise ValueError(
                "HiRISE Lookup Tables must have 1 or 256 entries."
            )

        if len(lut) == 1 and lut == [[0, 0]]:
            # A single-item (0, 0) LUT indicates that
            # no LUT has been applied to these data.
            self.table = False
        else:
            # The final two table entries must have identical values:
            for i in (-1, -2):
                if lut[i][0] != lut[i][1]:
                    text = {-1: "final", -2: "second to last"}
                    raise ValueError(
                        f"The {text[i]} element of the table must have "
                        f"identical values, and this does not: {lut[i]}"
                    )

            # Pre-compute the lookups:
            self.table = list()
            for i, pair in enumerate(lut):
                if pair[0] == pair[1] or i + 1 == len(lut):
                    self.table.append(pair[0])
                else:
                    self.table.append(
                        int(statistics.mean((lut[i][0], lut[i + 1][0])))
                    )

            # Ensure the first lut element is zero
            if lut[0][0] == 0:
                self.table[0] = 0
            else:
                raise ValueError(
                    f"The first LUT pair doesn't have a zero: {lut[0]}"
                )

        self.lut = lut

    def __str__(self):
        if self.table:
            return str(self.table)
        else:
            return "No applied LUT."

    def unlut(self, pixel: int) -> int:
        if self.table:
            return self.table[pixel]
        else:
            # That means there's a single pair "(0, 0)" which indicates
            # that no lut has been applied, so this function should no-op.
            return pixel

    def lookup(self, dn: int) -> int:
        if self.table:
            for i, pair in enumerate(self.lut):
                if pair[0] <= dn <= pair[1]:
                    return i
            else:
                return 255

        else:
            # See comment in unlut() above.
            return dn

    def specialpixels(self):
        """A kalasiris.specialpixels.SpecialPixels named tuple is
        returned which describes the values in this LUT.

        These are slightly different than the ISIS special pixel values,
        and ISIS special pixels values can't be used to deal with values
        that are read from these EDRs, however, those values are provided
        for your use by this function.
        """
        # Based on the FixDns8() and FixDns16() functions in
        # https://github.com/USGS-Astrogeology/ISIS3/blob/dev/isis/src/mro/apps/hi2isis/main.cpp
        if self.table:
            return isis.specialpixels.SpecialPixels(
                Min=1,
                Null=self.unlut(255),
                Lrs=0,
                Lis=0,
                His=self.unlut(254),
                Hrs=self.unlut(255),
                Max=self.lut[-3][1],
            )
        else:
            return isis.specialpixels.SpecialPixels(
                Min=1,
                Null=65535,
                Lrs=0,
                Lis=0,
                His=16383,
                Hrs=65535,
                Max=16382,
            )


def byte_info(name: str, label: dict) -> tuple:
    if "SAMPLE_BITS" in label[name]:
        if label[name]["SAMPLE_BITS"] == 16:
            byte_width = 2
        elif label[name]["SAMPLE_BITS"] == 8:
            byte_width = 1
        else:
            raise ValueError(
                "Don't know how to handle '{}'".format(
                    label[name]["SAMPLE_BITS"]
                )
                + f"as a value of SAMPLE_BITS in {name}, should be 8 or 16."
            )
    else:
        raise ValueError("Can't determine how to set the width.")

    if label[name]["SAMPLE_TYPE"].startswith("MSB"):
        byte_order = "big"
    else:
        byte_order = "little"

    if "UNSIGNED" in label[name]["SAMPLE_TYPE"]:
        byte_signed = False
    else:
        byte_signed = True

    return byte_width, byte_order, byte_signed


def object_asarray(file_path: os.PathLike, name: str) -> np.ndarray:
    """Return the data object of *name* from *file_path* as a numpy array.

    Data objects are found by having both a a pointer parameter (like ^*name*)
    and an object in the label with *name*.

    Doesn't work for objects that have columns.  Since such objects don't
    have SAMPLE_BITS and SAMPLE_TYPE fields, this will result in a KeyError.
    """
    label = pvl.load(str(file_path))
    width, b_order, b_signed = byte_info(name, label)

    lut = LUT_Table(
        label["INSTRUMENT_SETTING_PARAMETERS"]["MRO:LOOKUP_CONVERSION_TABLE"]
    )

    table = list()
    with open(file_path, mode="rb") as f:
        # Since The first byte is location 1 (not 0) in PDS-counting.
        f.seek(int(label[f"^{name}"].value) - 1)
        for line in range(label[name]["LINES"]):
            row = list()
            # Skip over the prefix bytes.
            f.seek(label[name]["LINE_PREFIX_BYTES"], os.SEEK_CUR)
            for samp in range(label[name]["LINE_SAMPLES"]):
                dn = lut.unlut(
                    int.from_bytes(
                        f.read(width), byteorder=b_order, signed=b_signed
                    )
                )
                row.append(dn)
            table.append(row)
            # Skip over the suffix bytes.
            f.seek(label[name]["LINE_SUFFIX_BYTES"], os.SEEK_CUR)

    return np.array(table)


def overwrite_object(file_path: os.PathLike, name: str, arr: np.ndarray):
    """The file at *file_path* will have its *name* object overwritten
    with the data in *array*.

    Doesn't work for objects that have columns.  Since such objects don't
    have SAMPLE_BITS and SAMPLE_TYPE fields, this will result in a KeyError.
    """
    label = pvl.load(str(file_path))
    width, b_order, b_signed = byte_info(name, label)

    if arr.shape != (label[name]["LINES"], label[name]["LINE_SAMPLES"]):
        raise ValueError(
            f"The object {name} has different dimensions "
            f"({label[name]['LINES']}, "
            f"{label[name]['LINE_SAMPLES']})than "
            f"the array ({arr.shape})."
        )

    lut = LUT_Table(
        label["INSTRUMENT_SETTING_PARAMETERS"]["MRO:LOOKUP_CONVERSION_TABLE"]
    )

    with open(file_path, mode="r+b") as f:
        # Since The first byte is location 1 (not 0) in PDS-counting.
        f.seek(int(label[f"^{name}"].value) - 1)
        for line in range(arr.shape[0]):
            # Skip over the prefix bytes.
            f.seek(label[name]["LINE_PREFIX_BYTES"], os.SEEK_CUR)
            for samp in range(arr.shape[1]):
                f.write(
                    lut.lookup(arr[line][samp]).to_bytes(
                        width, byteorder=b_order, signed=b_signed
                    )
                )
            # Skip over the suffix bytes.
            f.seek(label[name]["LINE_SUFFIX_BYTES"], os.SEEK_CUR)
    return



#!/usr/bin/env python
"""Converts LIS values in reverse-clock and buffer areas to
derived values.

For a variety of reasons, but highlighted with the August 2020
change of ADC settings, the HiRISE LUT process can clip the DN
histogram resulting in good data in the image area, but mostly LIS
values in the reverse-clock and buffer areas, which are crucial for
calibration.  This program attempts to replace "missing" data in those
areas with values derived from the ramp area, the dark area, and models
of expected reverse-clock medians.
"""

# Copyright 2020-2021, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Parts of this were inspired by fix_74.pro, find_zero2REV.pro (Dec 2020),
# and fix_74_all_new.pro (Feb 2021) by Alan Delamere, but the implementation
# here was written from scratch.


import argparse
import logging
import os
import shutil
import sys
from collections import abc
from pathlib import Path

import numpy as np
import scipy.signal as signal

import pvl
import kalasiris as isis

import hiproc.util as util
from hiproc.bitflips import apply_special_pixels
from hiproc.hirise import get_ChannelID_fromfile

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[util.parent_parser()],
    )
    parser.add_argument(
        "-o", "--output", required=False, default=".lisfix.cub"
    )
    parser.add_argument("file", help="ISIS Cube file.")
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    out_p = util.path_w_suffix(args.output, args.file)

    with util.main_exceptions(args.verbose):
        fixed = fix(
            args.file,
            out_p,
        )

    if not fixed:
        sys.exit(f"{args.file} did not need lisfix cleaning.")

    return


def fix(
    in_path: os.PathLike,
    out_path: os.PathLike,
    tolerance=0.4
) -> bool:
    """The ISIS cube file at *out_path* will be the result of running lis-fix
    processing of the file at *in-path*, if allowed by *tolerance*.  If the
    *tolerance* setting would result in the application of lis-fix, then
    no file will be created at *out_path* and this function will return
    False.  Otherwise there will be a file there, and this function
    will return True.

    If the fraction of LIS pixels in the Reverse-Clock areas is
    greater than *tolerance*, then all LIS pixels in the rev-clock
    will be converted to a real DN value based on modeling the slope
    of the ramp area.

    If the fraction of LIS pixels in the Buffer Pixel area is greater
    than *tolerance*, this algorithm will attempt to fix those LIS pixels,
    as well.

    If the file at *in_path* isn't an ISIS cube, a ValueError will be raised.

    This function anticipates a HiRISE cube that has the following
    table objects in its label: HiRISE Calibration Ancillary,
    HiRISE Calibration Image, HiRISE Ancillary.  If they are not present
    a KeyError will be raised.

    The HiRISE Calibration Ancillary table contains the BufferPixels
    and DarkPixels from either side of the HiRISE Calibration Image.
    The HiRISE Calibration Image table contains the Reverse-Clock,
    Mask, and Ramp Image areas.  The HiRISE Ancillary table contains
    the BufferPixels and DarkPixels from either side of the main
    Image Area.
    """

    in_p = Path(in_path)
    out_p = Path(out_path)
    fixed = False

    label = pvl.load(in_p)
    if "IsisCube" not in label:
        raise ValueError(f"The file at {in_p} is not an ISIS Cube.")

    binning = label["IsisCube"]["Instrument"]["Summing"]
    specialpix = getattr(
        isis.specialpixels, label["IsisCube"]["Core"]["Pixels"]["Type"]
    )

    # Get Rev-Clock data
    t_name = "HiRISE Calibration Image"
    hci_dict = isis.cube.get_table(in_path, t_name)
    cal_vals = np.array(hci_dict["Calibration"])

    cal_image = np.ma.masked_outside(
        cal_vals, specialpix.Min, specialpix.Max
    )

    mask_lines = int(20 / binning)
    ramp_start = 20 + mask_lines

    # Get the average slope of the dark ramp
    hca_dict = isis.cube.get_table(in_path, "HiRISE Calibration Ancillary")
    caldark_vals = np.array(hca_dict["DarkPixels"])
    caldark = np.ma.masked_outside(
        caldark_vals, specialpix.Min, specialpix.Max
    )
    # print(caldark.shape)
    dark_slopes = np.ma.apply_along_axis(
        get_ramp_slope, 0, caldark[ramp_start:, 1:]
    )
    # print(dark_slopes.shape)
    # print(dark_slopes)

    # CORRECT FOR MASKED LINES
    # ASSUME SLOPE IS DEFINED BY DARK COLUMNS
    dark_slope_mean = np.mean(dark_slopes)
    logger.info(f"Dark Ramp mean slope: {dark_slope_mean}")
    # FOR # J = 12, SZ(1) - 1 # DO
    # zero.line(j) = zero.line(j) - SL_D * (
    #         ystart(j) + 1 + 20. / info.bin * 103. / 89)
    zero_correction = dark_slope_mean * (20 / binning * 103 / 89)

    revclk_lisfrac = np.ma.count_masked(
        cal_image[:20, :]
    ) / cal_image[:20, :].size
    logger.info(
        f"Fraction of LIS pixels in Reverse-Clock area: {revclk_lisfrac:.2}"
    )
    if revclk_lisfrac > tolerance:
        fixed = True
    else:
        logger.info(
            f"Less than tolerance ({tolerance}), will not fix Reverse-Clock."
        )

    # Get Buffer Data
    calbuf_vals = np.array(hca_dict["BufferPixels"])
    calbuf = np.ma.masked_outside(
        calbuf_vals, specialpix.Min, specialpix.Max
    )

    ha_dict = isis.cube.get_table(in_path, "HiRISE Ancillary")
    dark_vals = np.array(ha_dict["DarkPixels"])
    dark = np.ma.masked_outside(
        dark_vals, specialpix.Min, specialpix.Max
    )
    # print(dark.shape)
    buffer_vals = np.array(ha_dict["BufferPixels"])
    buffer = np.ma.masked_outside(
        buffer_vals, specialpix.Min, specialpix.Max
    )
    # print(buffer)
    # print(buffer.dtype)
    # print(buffer.shape)

    buffer_lisfrac = np.ma.count_masked(buffer) / buffer.size
    logger.info(
        f"Fraction of LIS pixels in Buffer area: {buffer_lisfrac:.2}"
    )
    if buffer_lisfrac > tolerance:
        fixed = True
    else:
        logger.info(
            f"Less than tolerance ({tolerance}), will not fix Buffer."
        )

    ramp_lisfrac = np.ma.count_masked(
        cal_image[ramp_start:]
    ) / cal_image[ramp_start:].size

    if ramp_lisfrac > tolerance:
        fixed = False
        logger.warning(
            f"The fraction of LIS pixels in the ramp area ({ramp_lisfrac}) "
            f"is greater than the tolerance ({tolerance}), which prevents "
            f"lisfix processing."
        )

    if fixed:
        shutil.copy(in_path, out_path)
    else:
        return False

    # Fix rev-clock
    if revclk_lisfrac > tolerance:
        rev_model = ramp_rev_clock(
            cal_image[ramp_start:],
            label["IsisCube"]["Instrument"]["ChannelNumber"],
            zero_correction
        ).astype(int)
        # # Mask out all of the rev-clock for testing, so we get wholesale
        # # replacement
        # orig_cal_image = np.ma.copy(cal_image)
        # o_mean = np.ma.mean(orig_cal_image[:20, :], axis=0)
        # y_err_upper = np.ma.max(orig_cal_image[:20, :], axis=0) - o_mean
        # y_err_lower = o_mean - np.ma.min(orig_cal_image[:20, :], axis=0)

        # cal_image[:20, :] = np.ma.masked

        # # Some plotting for debugging:
        # import matplotlib.pyplot as plt
        # plt.plot(o_mean, 'k', label='Mean of Rev-Clock columns')
        # # plt.errorbar(
        # #     np.linspace(0, o_mean.size, num=o_mean.size, endpoint=False),
        # #     o_mean,
        # #     yerr=[y_err_lower, y_err_upper],
        # #     fmt='k',
        # #     label='Mean of Rev-Clock columns'
        # # )
        # for i in range(19):
        #     plt.plot(orig_cal_image[i, :], 'k.')
        # plt.plot(orig_cal_image[19, :], 'k.', label='Original Rev-Clock')
        # plt.plot(rev_model, 'r', label="Fixed")
        # plt.legend(loc='best')
        # plt.show()
        # sys.exit()

        # When we "fix" pixels, we must ensure that we are only fixing the LIS
        # pixels.  The cal_image has all special pixels masked out, so
        # we need to create a new structure that only masks the LIS pixels.
        cal_lismasked = np.ma.masked_equal(cal_vals, specialpix.Lis)
        logger.info("Fixing Reverse-Clock LIS pixels.")
        fixed_cal = np.ma.apply_along_axis(
            fix_rev_clock, 0, np.ma.concatenate((
                cal_lismasked,
                rev_model.reshape((1, rev_model.size))
            ))
        )

        logger.info("Writing out Reverse-Clock pixels.")
        hci_dict["Calibration"] = apply_special_pixels(
            fixed_cal, specialpix
        ).data.tolist()
        isis.cube.overwrite_table(out_p, t_name, hci_dict)

    # Fix the buffer area:
    if buffer_lisfrac > tolerance:
        # first_im_line = 19+(20+label["IsisCube"]["Instrument"]["Tdi"])/binning
        sp_frac = np.ma.count_masked(calbuf[:20, :]) / calbuf[:20, :].size
        if sp_frac <= tolerance:
            refr = int(np.ma.median(calbuf[:20, :]))
            logger.info(
                f"Median of first 20 lines of Calibration Buffer Pixels: {refr}"
            )
        else:
            logger.info(
                f"More than {tolerance:.1%} of the Calibration Buffer "
                f"Pixels have special pixel values: {sp_frac:.2%}."
            )
            isp = pvl.loads(
                isis.catoriglab(in_path).stdout
            )["INSTRUMENT_SETTING_PARAMETERS"]
            adc = label["IsisCube"]["Instrument"]["ADCTimingSetting"]
            if adc == -9999:
                adc = isp["MRO:ADC_TIMING_SETTINGS"]

            cid = get_ChannelID_fromfile(in_path)
            refr = int(
                revclock_model(
                    cid.ccdname + cid.ccdnumber + "_" + cid.channel,
                    binning,
                    label[
                        "IsisCube"
                    ]["Instrument"]["FpaPositiveYTemperature"].value,
                    adc
                )
            )
            logger.info(f"Using a model-based substitute ({refr}).")
            # import hiproc.img as img
            # lut = img.LUT_Table(isp["MRO:LOOKUP_CONVERSION_TABLE"])
            # if refr <= lut.table[1]:
            #     logger.error(f"Using a model-based substitute ({refr}).")
            # else:
            #     logger.error(
            #         f"Model ({refr}) was greater than LUT floor "
            #         f"({lut.table[1]}). Setting to LUT floor."
            #     )
            #     refr = lut.table[1]

        if np.ma.count_masked(dark[:20]) / dark[:20].size <= tolerance:
            refd = int(np.ma.median(dark[:20]))
            logger.info(f"Median of first 20 lines of Dark Pixels: {refd}")
        else:
            refd = int(np.ma.median(fixed_cal[:20, :]))
            logger.info(
                f"More than {tolerance:.1%} of the first 20 lines of "
                f"Dark Pixels have a real "
                f"value: {np.ma.count_masked(dark[:20]) / dark[:20].size} "
                f"Using the rev-clock median ({refd})."
            )

        model_buffer = dark[:, 2:14] - refd + refr
        # print(model_buffer)
        # print(model_buffer.dtype)
        # print(model_buffer.shape)

        # # Mask out all of the buffer for testing, so we get wholesale
        # # replacement
        # buffer.mask = True

        logger.info("Fixing Buffer pixels.")
        buffer_lismasked = np.ma.masked_equal(buffer_vals, specialpix.Lis)
        fixed_buf = np.ma.apply_along_axis(
            fix_buffer,
            1,
            np.ma.concatenate((buffer_lismasked, model_buffer), 1),
            buffer_lismasked.shape[1]
        )
        # print(fixed_buf)
        # print(fixed_buf.dtype)
        # print(fixed_buf.shape)

        ha_dict["BufferPixels"] = apply_special_pixels(
            fixed_buf, specialpix
        ).data.tolist()
        logger.info("Writing out Buffer pixels.")
        isis.cube.overwrite_table(out_p, "HiRISE Ancillary", ha_dict)

    return fixed


def fit_ramp(col: np.ma.array):
    """Returns the slope from the ramp pixels."""
    return np.ma.polyfit(np.arange(col.size), col, 1)


def get_ramp_slope(col: np.ma.array):
    return fit_ramp(col)[0]


def get_ramp_intercept(col: np.ma.array):
    return fit_ramp(col)[1]


def fix_rev_clock(
    cal_image_col: np.ma.array,
    lis_fraction=0.2,
):
    """Returns a np.masked array where the masked pixels in the
    *cal_image_col* are replaced with a fill value from the end
    of the column, if the percent of masked pixels is greater than or
    equal to *lis_fraction*.

    It is assumed that cal_image_col array has one more final entry
    (at position [-1]) than it should, and that value is used to
    fill the masked areas, if needed, and then is not included
    in the array on return.
    """
    if np.ma.count_masked(cal_image_col[:20]) / 20 >= lis_fraction:
        cal_image_col[:20] = cal_image_col[:20].filled(
            fill_value=cal_image_col[-1]
        )
    return cal_image_col[:-1]


def ramp_rev_clock(
    ramp_image: np.ma.array,
    chan: int,
    zero_correction=0,
):
    ramp_model = np.ma.apply_along_axis(
        get_ramp_intercept, 0, ramp_image
    )

    return flatten(ramp_model - zero_correction, chan)


def fix_buffer(
    buf_row: np.ma.array,
    stop,
    lis_fraction=0.5,
):
    """Returns a np.masked array the length of *buf_row[:stop]* with those
    same values, but with the masked pixels in the [:stop] part the row
    replaced with pixels from the [stop:] part, if the percent of masked
    pixels in [stop:] is greater than or equal to *lis_fraction*."""
    if np.ma.count_masked(buf_row[:stop]) / stop >= lis_fraction:
        # print(f"row: {buf_row[:stop]} {buf_row[stop:]}")
        # print(buf_row.data[:stop])
        for i in range(stop):
            # print(f"masks: {buf_row.mask[i]}, {buf_row.mask[i + stop]}")
            # print(f"mask type: {type(buf_row.mask[i])} ")
            # print(f"values: {buf_row[i]}, {buf_row[i + stop]}")
            # The mask values are really np.bool_ values, so compare
            # to ints, not Python True, False objects.
            if buf_row.mask[i] == 1 and buf_row.mask[i + stop] == 0:
                buf_row[i] = buf_row[i + stop]
            # print(f"values: {buf_row[i]}, {buf_row[i + stop]}")
        # print(buf_row.data[:stop])
        # print(buf_row[:stop])

    return buf_row[:stop]


def revclock_model(
    channelid: str,
    binning: int,
    temperature: float,
    adc: abc.Sequence
) -> float:
    # The slope and intercept data is from a presentation by Ken
    # Herkenhoff, in July 2020, entitled "Analysis of flight reverse-clocked
    # data" in the file Full\ reverse\ analysis.pptx
    # However, I think the ADC 7, 4 values were added after that date, but
    # sometime before Feb 2021.
    adc = tuple(adc)
    if adc == (5, 4):
        if binning != 1:
            raise KeyError(
                "There is no model for ADC 5,4 data with a binning other "
                "than 1."
            )
        pairs = {
            "RED0_0": (1278, 4.08),
            "RED1_0": (1072, 3.32),
            "RED2_0": (1214, 3.45),
            "RED3_0": (1083, 3.70),
            "RED4_0": (1124, 3.52),
            "RED5_0": (1148, 4.90),
            "RED6_0": (992, 4.14),
            "RED7_0": (1117, 1.90),
            "RED8_0": (1269, 4.56),
            "RED9_0": (1144, 3.33),
            "IR10_0": (1121, 4.92),
            "IR11_0": (1171, 3.35),
            "BG12_0": (986, 3.78),
            "BG13_0": (1027, 4.10),
            "RED0_1": (1341, 5.70),
            "RED1_1": (1114, 4.65),
            "RED2_1": (1093, 4.73),
            "RED3_1": (1187, 5.15),
            "RED4_1": (1078, 4.54),
            "RED5_1": (1335, 5.50),
            "RED6_1": (1154, 5.36),
            "RED7_1": (1355, 5.55),
            "RED8_1": (1398, 4.27),
            "RED9_1": (1255, 5.47),
            "IR10_1": (991, 5.81),
            "IR11_1": (1145, 4.20),
            "BG12_1": (1112, 5.67),
            "BG13_1": (974, 5.32),
        }
        intercept, slope = pairs[channelid]

    elif adc == (7, 4):
        pairs = {
            "RED1_0": {
                1: (1010.4, 0.7),
                2: (930.4, 1.9),
                4: (914.4, 2.8),
            },
            "RED1_1": {
                1: (869.2, 3.5),
                2: (845.3, 3.1),
                4: (699.2, 7.1),
            },
            "RED2_0": {
                1: (1051.2, 1.6),
                2: (962.7, 2.6),
                4: (923.6, 4.1),
            },
            "RED2_1": {
                1: (842.5, 3.9),
                2: (734.9, 5.1),
                4: (644.9, 7.7),
            },
            "RED3_0": {
                1: (958.8, 1.8),
                2: (841.5, 3.7),
                4: (788.1, 5.5),
            },
            "RED3_1": {
                1: (916.8, 3.5),
                2: (759.8, 6.1),
                4: (648.4, 9.1),
            },
        }
        intercept, slope = pairs[channelid][binning]

    else:
        raise KeyError(
            f"Do not have a model for {adc}."
        )

    return temperature * slope + intercept


def flatten(arr: np.array, chan: int):
    # The idea here is to flatten the zero_correction array between the pause
    # points
    sos = signal.butter(1, 0.5, output="sos")
    pl = list()
    for samp, width in zip(util.ch_pause[chan], util.ch_width[chan]):
        sl = util.pause_slicer(samp, width)
        pl.append(sl.start)
        pl.append(sl.stop)

    # # Supports debug plotting below, 2 lines
    # fixed = np.copy(arr)
    # filtered_whole = np.full_like(arr, np.mean(arr))
    for pslice in (
        slice(None, pl[0]),
        slice(pl[1], pl[2]),
        slice(pl[3], pl[4]),
        slice(pl[5], None)
    ):
        m = np.mean(arr[pslice])
        filtered = signal.sosfilt(sos, arr[pslice] - m)
        arr[pslice] = (arr[pslice] - m - filtered) + m
        # # Supports debug plotting below, 2 lines
        # fixed[pslice] = (arr[pslice] - m - filtered) + m
        # filtered_whole[pslice] = filtered + m

    # # Some plotting for debugging:
    # import matplotlib.pyplot as plt
    # plt.plot(arr, 'k', label='Original')
    # # plt.plot(filtered_whole, 'b', label='Filtered')
    # plt.plot(fixed, 'r', label="Fixed")
    # plt.legend(loc='best')
    # plt.show()
    # sys.exit()

    return arr


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Start from .img files and run them through a complete processing chain."""

# Copyright 2020-2021, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Need to think about endpoints, etc.
# Minimal 'run' from channels to single file is to stop at HiccdStitch
#
# If you knew you only wanted to stop at HiNoProj, could do this next,
# but in order to run HiPrecisionInit to decide HiNoProj or HiJACK, need
# slither files.
#
# For the remaining 'end points' these next three must be done:
#   - HiColorInit
#   - HiJitReg
#   - HiSlither
#
# "Color" end:
# Then run HiColorNorm and HiBeautify
#
# "Red" end:
# HiPrecisionInit, and then HiNoProj or HiJACK.
#
# So, stops are: HiccdStitch, "Color", and "red"


import argparse
import concurrent.futures
import itertools
import logging
import os
import pkg_resources
from pathlib import Path

import pvl.new as pvl

import hiproc.hirise as hirise
import hiproc.util as util
import hiproc.EDR_Stats as EDR_Stats
import hiproc.HiCal as HiCal
import hiproc.HiStitch as HiStitch
import hiproc.HiccdStitch as HiccdStitch
import hiproc.HiColorInit as HiColorInit
import hiproc.HiJitReg as HiJitReg
import hiproc.HiSlither as HiSlither
import hiproc.HiColorNorm as HiColorNorm
import hiproc.HiBeautify as HiBeautify
import hiproc.HiPrecisionInit as HiPrecisionInit
import hiproc.HiNoProj as HiNoProj
import hiproc.HiJACK as HiJACK

logger = logging.getLogger(__name__)


class ChannelCube(hirise.ChannelID):
    def __init__(self, pathlike, db=None):
        self.path = Path(pathlike)
        self.nextpath = self.path
        super().__init__(hirise.get_ChannelID_fromfile(self.path))
        self.db = db


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-c", "--color", action="store_true", help="Perform Color processing."
    )
    parser.add_argument(
        "-p",
        "--precision",
        action="store_true",
        help="Perform HiPrecision processing to result "
        "in NOPROJ.cub that might be processed through "
        "HiJACK.",
    )
    parser.add_argument(
        "-j",
        "--jack",
        action="store_true",
        help="Do not test if HiJACK is needed, but force HiJACK to run.",
    )
    parser.add_argument(
        "--max_workers",
        default=None,
        type=int,
        help="This program uses multiprocessing to parallelize some of the "
             "work.  This multiprocessing will default to use the number of "
             "processors on the host machine.  If you want to throttle this "
             "to use less resources on your machine, indicate the number of "
             "processors you want to use."
    )
    parser.add_argument(
        "img",
        metavar="some.img-file",
        nargs="+",
        help="More than one can be listed here.",
    )
    parser.add_argument(
        "--conf_dir",
        required=False,
        type=Path,
        default=Path(pkg_resources.resource_filename(__name__, "data/")),
        help="Directory where configuration files for each of the hiproc "
             "algorithms are.  Defaults to the configuration files "
             "distributed with the library."
    )
    return parser


def main():
    args = arg_parser().parse_args()

    util.set_logger(args.verbose, args.logfile, args.log)

    oid = hirise.get_ObsID_fromfile(args.img[0])
    parent = Path(args.img[0]).parent
    if args.img == 1:
        imgs = parent.glob(f"{oid}*.img") + parent.glob(f"{oid}*.IMG")
    else:
        imgs = args.img

    db_list = None
    with util.main_exceptions(args.verbose):
        try:
            get_cubes(f"{oid}*balance.cub", parent)
        except FileNotFoundError:
            chancubes = edr2stitch(
                imgs,
                args.conf_dir,
                max_workers=args.max_workers,
                keep=args.keep
            )
            db_list = [x.db for x in chancubes]

        if args.color:
            color(oid, args.conf_dir, parent, db_list, keep=args.keep)

        if args.precision or args.jack:
            jack = False
            if args.jack:
                jack = True
            else:
                slithers = list(parent.glob("*.slither.txt"))

                # HiPrecisionInit to determine if you need to HiNoProj or
                # HiJACK takes *slither.txt
                (HiJACK_flags, averages, thresh) = HiPrecisionInit.check(
                    slithers, pvl.load(args.conf_dir / "HiPrecisionInit.conf")
                )

                logger.info(HiPrecisionInit.message(
                    HiJACK_flags, averages, thresh, slithers
                ))

                if any(HiJACK_flags):
                    jack = True

            precision(oid, args.conf_dir, parent, hijack=jack, keep=args.keep)

    return


def get_cubes(glob: str, parent: Path) -> list:
    cubes = list(parent.glob(glob))
    if len(cubes) == 0:
        raise FileNotFoundError(
            f"Did not find any files that match {glob} in {parent}."
        )
    return cubes


def edr_cal(
    img: os.PathLike,
    out_edr: os.PathLike,
    out: os.PathLike,
    edr_conf: dict,
    hical_conf: dict,
    hical_conf_path: os.PathLike,
    bitflipwidth=0,
    lis_tolerance=1.0,
    keep=False,
) -> dict:
    # EDR_Stats
    db = EDR_Stats.EDR_Stats(
        img,
        out_edr,
        edr_conf,
        keep=keep,
    )

    # HiCal
    return HiCal.HiCal(
        out_edr,
        out,
        db,
        hical_conf,
        hical_conf_path,
        None,
        None,
        bitflipwidth,
        lis_tolerance,
        keep=keep,
    )


def edr2stitch(
    images,
    conf_dir,
    bitflipwidth=0,
    lis_tolerance=1.0,
    max_workers=None,
    keep=False,
):
    chids = list()
    edr_conf = pvl.load(conf_dir / "EDR_Stats_gains_config.pvl")
    hical_conf = HiCal.conf_setup(
        pvl.load(conf_dir / "HiCal.conf"),
        pvl.load(conf_dir / "NoiseFilter.conf"),
    )
    with concurrent.futures.ProcessPoolExecutor(
        max_workers=max_workers
    ) as executor:
        future_dbs = dict()
        for i in images:
            out_edr = util.path_w_suffix(".EDR_Stats.cub", i)
            out_hical = util.path_w_suffix(".HiCal.cub", out_edr)

            f = executor.submit(
                edr_cal,
                i,
                out_edr,
                out_hical,
                edr_conf,
                hical_conf,
                conf_dir / "HiCal.conf",
                bitflipwidth=bitflipwidth,
                lis_tolerance=lis_tolerance,
                keep=False,
            )
            future_dbs[f] = out_hical

        for future in concurrent.futures.as_completed(future_dbs):
            if future.exception() is None:
                out_hical = future_dbs[future]
                chids.append(ChannelCube(out_hical, future.result()))
            elif isinstance(future.exception(), UserWarning):
                logger.warning(future.exception())

    # HiStitch
    # get Channel pairs
    cids = list()
    stitch_conf = pvl.load(conf_dir / "HiStitch.conf")
    with concurrent.futures.ProcessPoolExecutor(
        max_workers=max_workers
    ) as executor:
        future_tuples = list()
        for chid1, chid2 in get_CCDpairs(chids):
            f = executor.submit(
                # (db, o_path) = HiStitch.HiStitch(
                HiStitch.HiStitch,
                chid1.nextpath,
                chid2.nextpath if chid2 is not None else None,
                chid1.db,
                chid2.db if chid2 is not None else None,
                ".HiStitch.cub",
                stitch_conf,
                keep=keep,
            )
            future_tuples.append(f)

        for future in concurrent.futures.as_completed(future_tuples):
            (db, o_path) = future.result()
            cid = HiccdStitch.HiccdStitchCube(o_path)
            cid.gather_from_db(db)
            cids.append(cid)

    # HiccdStitch, makes balance cubes
    # need to separate by color:
    color_groups = get_color_groups(cids)
    for color_group in color_groups.values():
        db, out_stitch = HiccdStitch.HiccdStitch(
            color_group,
            ".HiccdStitch.cub",
            pvl.load(conf_dir / "HiccdStitch.conf"),
            sline=None,
            eline=None,
            keep=keep,
        )
    # HiColorInit
    #   takes *balance.cub
    #   creates *[IR|BG]*.balance.precolor.cub
    #   Can then run JitPlot on these *.balance.precolor.cub
    HiColorInit.HiColorInit(
        [c.nextpath for c in cids], ".precolor.cub", keep=keep
    )

    # HiJitReg
    #   takes tmp/*balance.cub tmp/*balance.precolor.cub
    #   creates *regdef.pvl and *flat.tab files
    for_jitreg = list()
    for color, balcubes in color_groups.items():
        if color == "RED":
            for c in balcubes:
                for_jitreg.append(c.nextpath)
        else:
            for c in balcubes:
                for_jitreg.append(c.nextpath.with_suffix(".precolor.cub"))

    HiJitReg.HiJitReg(
        for_jitreg, pvl.load(conf_dir / "HiJitReg.conf"), keep=keep
    )

    # HiSlither
    #   takes same as HiJitReg (and assumes its products are available.
    #   creates *slither.txt, *slither.cub, and *COLOR[4|5].cub
    #   Can then run SlitherStats on the *slither.txt
    HiSlither.HiSlither(for_jitreg)

    return chids


def color(obsid, conf_dir: Path, parent: Path, db_list: list, keep=False):
    colors = get_cubes(f"{obsid}_COLOR*.cub", parent)

    # HiColorNorm - only for color
    #   takes *COLOR[4|5].cub
    #   creates *UNFILTERED_COLOR[4|5].cub and *COLOR[4|5].HiColorNorm.cub
    HiColorNorm.HiColorNorm(
        colors,
        "_COLOR.cub",
        pvl.load(conf_dir / "HiColorNorm.conf"),
        db_list=db_list,
        keep=keep,
    )

    # HiBeautify - only for color
    #   takes tmp/*.HiColorNorm.cub
    #   creates *IRB.cub and *RGB.cub
    for x in colors:
        x.with_suffix(".HiColorNorm.cub")
    HiBeautify.HiBeautify(
        [x.with_suffix(".HiColorNorm.cub") for x in colors],
        pvl.load(conf_dir / "HiBeautify.conf"),
    )
    return


def precision(obsid, conf_dir: Path, parent: Path, hijack=False, keep=False):

    if hijack:
        bal_cubs = get_cubes(f"{obsid}*balance.cub", parent)
        # HiJACK - starts with balance cubes, needs all the colors
        #   takes all color balance.cubs
        #   creates a whole slow of files in HiJACK/
        HiJACK.HiJACK(
            bal_cubs,
            conf_dir,
            outdir=(parent / "HiJACK"),
            plot=False,
            keep=keep,
        )
    else:
        red_bal_cubs = get_cubes(f"{obsid}_RED*balance.cub", parent)
        # HiNoProj - alternate to HiccdStitch, starts with balance cubes
        #   takes only REDS: *RED*balance.cub
        #   creates PSP_010502_2090_RED.NOPROJ.cub
        HiNoProj.HiNoProj(
            red_bal_cubs, pvl.load(conf_dir / "HiNoProj.conf"), keep=keep
        )

    return


def get_CCDpairs(chids: list) -> list:
    """Return a list of two-tuples which contain pairs of ChannelCubes
    that belong to the same CCD from *chids*.
    """
    chids.sort()
    pairs = list()
    for k, g in itertools.groupby(chids, lambda x: x.ccdnumber):
        g_list = list(g)
        if len(g_list) > 2:
            raise ValueError(
                f"The list of channel IDs ({chids}) has more than two channels "
                f"from the same CCD, which shouldn't be possible, are they "
                f"from different observations?"
            )
        elif len(g_list) == 1:
            g_list.append(None)

        pairs.append(g_list)

    return pairs


def get_color_groups(cids: list) -> list:
    """Return a list of lists where each list contains
    ChannelCubes that are from the same COLOR of detector
    from *cids*.
    """
    cids.sort()
    color_groups = dict()
    for k, g in itertools.groupby(cids, lambda x: x.ccdname):
        color_groups[k] = list(g)

    return color_groups



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `utils` module."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
import unittest
from .utils import resource_check as rc


class TestResources(unittest.TestCase):
    def setUp(self):
        t = Path("tests")
        self.exists1 = t / "test_hirise.py"
        self.exists2 = t / "test_utils.py"
        self.not1 = t / "notexists" / "foo"
        self.not2 = t / "notexists" / "foo2"

    def test_have(self):
        (truth, test) = rc(self.exists1, self.exists2)
        self.assertEqual(truth, test)

    def test_have_none(self):
        (truth, test) = rc(self.not1, self.not2)
        self.assertNotEqual(truth, test)

    def test_have_some(self):
        (truth, test) = rc(self.exists1, self.not1)
        self.assertNotEqual(truth, test)



#!/usr/bin/env python
"""This module has tests for the HiRISE HiJACK functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pathlib import Path
from unittest.mock import call

# from unittest.mock import mock_open
from unittest.mock import patch
from unittest.mock import Mock

# import hiproc.hirise as hirise
import hiproc.HiJitReg as hjr
import hiproc.HiNoProj as hnp
import hiproc.HiJACK as hjk

hijackconf_path = Path("data") / "HiJACK.conf"
resjitconf_path = Path("data") / "ResolveJitter.conf"


def getkey(cube, group, key):
    values = {
        "ObservationId": "PSP_010502_2090",
        "ProductId": None,
        "Summing": 2,
        "Lines": 1024,
        "Samples": 1024,
        "Bands": 3,
        "TDI": 64,
        "Center": "(900, 700, 500) <NANOMETERS>",
        "SourceProductId": ("PSP_010502_2090_RED4_0, PSP_010502_2090_RED4_1"),
        "StitchedProductIds": (
            "PSP_010502_2090_RED4_0, PSP_010502_2090_RED4_1"
        ),
    }
    return values[key]


class TestHiJACK(unittest.TestCase):
    @patch("hiproc.HiNoProj.isis.getkey_k", side_effect=getkey)
    def setUp(self, getkey_k):
        self.r4 = hnp.Cube("dummy/PSP_010502_2090_RED4.HiStitch.balance.cub")
        self.r3 = hnp.Cube("dummy/PSP_010502_2090_RED3.HiStitch.balance.cub")
        self.r5 = hnp.Cube("dummy/PSP_010502_2090_RED5.HiStitch.balance.cub")
        self.i1 = hnp.Cube("dummy/PSP_010502_2090_IR10.HiStitch.balance.cub")

        self.sets = dict(
            CCDs="RED3,RED4,RED5,BG12",
            Alt_1_CCDs="RED4,RED5,RED6,IR11",
            Alt_2_CCDs="RED3,RED4,RED5,IR10",
            Alt_3_CCDs="IR10,RED4,RED5,BG12",
        )

    @patch("hiproc.HiJACK.shutil.move")
    @patch("hiproc.HiJACK.isis.handmos")
    @patch("hiproc.HiJACK.isis.hijitreg")
    @patch("hiproc.HiJACK.shutil.copy")
    @patch("hiproc.HiJACK.isis.reduce")
    @patch("hiproc.HiJACK.isis.enlarge")
    def test_match_red(
        self, m_enlarge, m_reduce, m_copy, m_hjr, m_hand, m_move
    ):

        cubes = [self.r4, self.r3, self.r5, self.i1]
        for c in cubes:
            c.next_path = c.path.with_suffix(".next.cub")
        cubes.sort()

        hjk.match_red(cubes, self.r5, "dummy/flat")
        m_enlarge.assert_not_called()
        m_reduce.assert_called_once()
        self.assertEqual(3, m_copy.call_count)
        self.assertEqual(
            m_hjr.call_args_list,
            [
                call(
                    self.r4.next_path,
                    flat="dummy/flat",
                    match=self.r5.next_path,
                    row=20,
                )
            ],
        )

    def test_find_common(self):
        s1 = set(("RED3", "RED4", "RED5", "BG12"))
        s2 = set(("RED4", "RED5", "RED6", "IR11"))
        s3 = set(("RED3", "RED4", "RED5", "IR10"))
        s4 = set(("IR10", "RED4", "RED5", "BG12"))

        self.assertEqual("RED4", hjk.find_common(s1))
        self.assertEqual("RED5", hjk.find_common(s2))
        self.assertEqual("RED4", hjk.find_common(s3))
        self.assertEqual("RED4", hjk.find_common(s4))
        self.assertRaises(ValueError, hjk.find_common, set(("RED1", "RED2")))

    def test_ccd_set_generator(self):
        g = hjk.ccd_set_generator(self.sets)
        self.assertEqual(next(g), {"BG12", "RED3", "RED5", "RED4"})
        self.assertEqual(next(g), {"IR11", "RED6", "RED5", "RED4"})
        self.assertEqual(next(g), {"IR10", "RED3", "RED5", "RED4"})
        self.assertEqual(next(g), {"BG12", "IR10", "RED5", "RED4"})
        self.assertRaises(StopIteration, next, g)

    def test_determine_resjit_set(self):
        cubes = [self.r4, self.r3, self.r5, self.i1]
        cubes.sort()

        (rcubes, common) = hjk.determine_resjit_set(
            cubes, hjk.ccd_set_generator(self.sets)
        )
        self.assertEqual(self.r4, common)
        rcubes.sort()
        self.assertEqual(cubes, rcubes)

    @patch("hiproc.HiJitReg.run_HiJitReg")
    @patch("hiproc.HiNoProj.isis.getkey_k", side_effect=getkey)
    def test_make_flats(self, m_get, m_rhjr):
        cubes = [self.r4, self.r3, self.r5, self.i1]
        for c in cubes:
            c.next_path = c.path.with_suffix(".next.cub")
        cubes.sort()

        conf = {
            "AutoRegistration": {
                "ControlNet": {
                    "Control_Cols_Red": 1,
                    "Control_Cols_Color": 3,
                    "Control_Lines": 20,
                },
                "Algorithm": {"Tolerance": 0.7, "Increment": 0.1, "Steps": 2},
                "PatternChipRed": {"Samples": 17, "Lines": 45},
                "PatternChipColor": {"Samples": 100, "Lines": 60},
                "SearchChipRed": {"Samples": 30, "Lines": 70},
                "SearchChipColor": {"Samples": 140, "Lines": 100},
                "SearchLongerChip": {"Samples": 130, "Lines": 140},
                "AnaylyzeFlat": {"Minimum_Good": 0.20},
            },
            "Smoothing": {
                "Exclude_Limit": 2,
                "Badness_Limit": 1,
                "Boxcar_Length": 10,
            },
        }

        with patch("hiproc.HiJitReg.JitterCube") as m_jit:
            m_path = Mock(spec_set=Path)
            m_jit().flattab_path = m_path
            s = hjk.make_flats(cubes, self.r5, conf, "tt", keep=True)
            self.assertEqual(s, [m_path] * (len(cubes) - 1))

        with patch("hiproc.HiJitReg.Analyze_Flat", return_value=1):
            s = hjk.make_flats(cubes, self.r5, conf, "tt", keep=True)
            flat_list = list()
            for x in (self.r3, self.r4, self.i1):
                p = hjr.JitterCube.get_pair_name(x, self.r5.get_ccd())
                flat_list.append(self.r5.path.parent / (p + ".flat.tab"))
            self.assertEqual(flat_list, s)

    @unittest.skip("ResolveJitter needs more work")
    @patch("hiproc.HiJACK.subprocess.run")
    def test_ResolveJitter(self, m_run):
        cubes = [self.r4, self.r3, self.r5, self.i1]
        for c in cubes:
            c.next_path = c.path.with_suffix(".next.cub")
        cubes.sort()

        flat_list = list()
        for x in (self.r3, self.r4, self.i1):
            p = hjr.JitterCube.get_pair_name(x, self.r5.get_ccd())
            flat_list.append(self.r5.path.parent / (p + ".flat.tab"))

        with patch(
            "hiproc.HiJACK.determine_resjit_set", return_value=(cubes, self.r5)
        ):
            with patch("hiproc.HiJACK.make_flats", return_value=flat_list):
                m_path = Mock(spec_set=Path)
                m_path.parent = self.r5.path.parent
                conf = dict(
                    ResolveJitter=self.sets,
                    AutoRegistration={"ControlNet": {"Control_Lines": 10}},
                )
                hjk.ResolveJitter(cubes, conf, m_path, "tt", keep=True)
                rjp = "/Users/rbeyer/software/HiPrecision_Oleg/resolveJitter"
                self.assertEqual(
                    m_run.call_args_list,
                    [
                        call(
                            [
                                rjp,
                                str(m_path.parent),
                                str(self.r5.get_obsid()),
                                "10",
                                flat_list[0].relative_to(m_path.parent),
                                "-1",
                                flat_list[1].relative_to(m_path.parent),
                                "-1",
                                flat_list[2].relative_to(m_path.parent),
                                "-1",
                            ],
                            check=True,
                        )
                    ],
                )

    @patch("hiproc.HiNoProj.handmos_side")
    @patch("hiproc.HiNoProj.fix_labels")
    def test_mosaic_dejittered(self, m_fix, m_hand):
        out_p = "output_path"
        prodid = "product_id"
        self.assertRaises(
            ValueError,
            hjk.mosaic_dejittered,
            [self.r4, self.r3, self.r5],
            out_p,
            prodid,
        )
        hjk.mosaic_dejittered([self.r4, self.r5], out_p, prodid)
        self.assertEqual(
            m_hand.call_args_list,
            [call([self.r4, self.r5], self.r5, out_p, left=True)],
        )
        self.assertEqual(
            m_fix.call_args_list,
            [call([self.r4, self.r5], out_p, str(self.r5), prodid)],
        )

    @patch("hiproc.HiNoProj.isis.getkey_k", side_effect=getkey)
    @patch("hiproc.HiJACK.Path.mkdir")
    @patch("hiproc.HiJACK.ResolveJitter")
    @patch("hiproc.HiNoProj.fix_labels")
    @patch("hiproc.HiJACK.isis.PathSet")
    @patch("hiproc.HiJACK.plot_flats")
    @patch("hiproc.HiJACK.isis.fromlist.temp")
    @patch("hiproc.HiJACK.isis.cubeit")
    @patch("hiproc.HiJACK.isis.handmos")
    @patch("hiproc.HiJACK.mosaic_dejittered")
    @patch("hiproc.HiJACK.shutil.copyfile")
    @patch("hiproc.HiJACK.isis.hijitter")
    @patch("hiproc.HiNoProj.copy_and_spice")
    @patch("hiproc.HiNoProj.conf_check")
    @patch("hiproc.HiJACK.pvl.load")
    @patch("hiproc.HiJACK.match_red")
    def test_HiJACK(
        self,
        m_match,
        m_pvl,
        m_ccheck,
        m_cns,
        m_hijit,
        m_copy,
        m_mosdejit,
        m_handmos,
        m_cubeit,
        m_fromlist,
        m_plot,
        m_PathSet,
        m_fixlab,
        m_ResJit,
        m_mkdir,
        m_getkey
    ):
        confd = "confdir"
        outd = "outdir"
        prehijack_flat = Path("RED5-RED4flat1")
        with patch(
            "hiproc.HiNoProj.add_offsets",
            side_effect=[
                (
                    [self.r4, self.r5],
                    [prehijack_flat, Path("RED3-RED4flat2")],
                ),
                ([self.i1, self.r3], ["irflat1", "irflat2"]),
                ([self.i1, self.r3], ["bgflat1", "bgflat2"]),
            ],
        ):
            hjk.HiJACK(
                [
                    Path("dummy/PSP_010502_2090_RED3.HiStitch.balance.cub"),
                    Path("dummy/PSP_010502_2090_RED4.HiStitch.balance.cub"),
                    Path("dummy/PSP_010502_2090_RED5.HiStitch.balance.cub"),
                    Path("dummy/PSP_010502_2090_IR10.HiStitch.balance.cub")
                ],
                confd,
                outd,
                base_ccd_number=4,
                plot=False,
                keep=True,
            )

            m_match.assert_called_once()
            self.assertEqual(
                m_pvl.call_args_list,
                [
                    call(confd + "/ResolveJitter.conf"),
                    call(confd + "/HiJACK.conf"),
                ],
            )
            m_ccheck.assert_called_once()
            self.assertEqual(m_cns.call_count, 4)
            m_hijit.assert_called_once()

            oid = str(self.r5.get_obsid())
            r5_p = Path(
                f"{outd}/"
                + self.r5.path.with_suffix("")
                .with_suffix(".dejittered.cub")
                .name
            )
            r3_p = Path(
                f"{outd}/"
                + self.r3.path.with_suffix("")
                .with_suffix(".dejittered.cub")
                .name
            )
            rnp = Path(f"{outd}/{oid}_RED.NOPROJ.cub")
            r45np = Path(f"{outd}/{oid}_RED4-5.NOPROJ.cub")
            irnp = Path(f"{outd}/{oid}_IR.NOPROJ.cub")
            bgnp = Path(f"{outd}/{oid}_BG.NOPROJ.cub")
            phjf = Path(f"{outd}/{oid}.RED5-RED4_posthijack.flat.tab")
            self.assertEqual(
                m_copy.call_args_list,
                [
                    call(r5_p, rnp),
                    call(None, r45np),
                    call(None, irnp),
                    call(None, bgnp),
                    call(prehijack_flat, phjf),
                ],
            )

            self.assertEqual(
                m_mosdejit.call_args_list,
                [
                    call([self.r4, self.r5], r45np, f"{oid}_RED4-5"),
                    call([self.i1, self.r3], irnp, f"{oid}_IR"),
                    call([self.i1, self.r3], bgnp, f"{oid}_BG"),
                ],
            )

            r4_p = Path(
                f"{outd}/"
                + self.r4.path.with_suffix("")
                .with_suffix(".dejittered.cub")
                .name
            )
            self.assertEqual(
                m_handmos.call_args_list,
                [
                    call(
                        None,
                        mosaic=rnp,
                        outband=1,
                        outline=1,
                        outsample=1,
                        priority="ontop",
                    )
                ],
            )

            irbnp = Path(f"{outd}/{oid}_IRB.NOPROJ.cub")
            self.assertEqual(m_cubeit.call_args_list[0][1]["proplab"], r45np)
            self.assertEqual(m_cubeit.call_args_list[0][1]["to"], irbnp)

            i1_p = Path(
                f"{outd}/"
                + self.i1.path.with_suffix("")
                .with_suffix(".dejittered.cub")
                .name
            )

            self.assertEqual(
                m_fromlist.call_args_list,
                [
                    call(
                        [
                            self.r3.path,
                            self.r4.path,
                            self.r5.path,
                            self.i1.path,
                        ]
                    ),
                    call([r3_p, r4_p, r5_p, i1_p]),
                    call([irnp, r45np, bgnp]),
                ],
            )

            m_plot.called_once()
            m_PathSet.called_once()

            self.assertEqual(
                m_fixlab.call_args_list,
                [call([self.r4, self.r5], rnp, self.r5, f"{oid}_RED")],
            )

            self.assertEqual(
                m_ResJit.call_args_list[0][0][0],
                [self.r3, self.r4, self.r5, self.i1],
            )
            self.assertEqual(
                m_ResJit.call_args_list[0][0][2],
                # Path(f"{outd}/{oid}_jitter_cpp.txt"),
                Path(f"{outd}/{oid}_jitter_py.txt"),
            )



#!/usr/bin/env python
"""This module contains HiRISE utility functions."""

# Copyright 2020, Ross A. Beyer (rbeyer@seti.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import subprocess
from itertools import repeat
from pathlib import Path

import kalasiris as isis

phase_names = (
    "INT",
    "CAL",
    "ATL",
    "KSC",
    "LAU",
    "CRU",
    "APR",
    "AEB",
    "TRA",
    "PSP",  # 'REL'
    "ESP",
)

phase_max_orbit = dict(
    zip(phase_names, repeat(0)), TRA=1000, PSP=11248, ESP=None
)

ccd_numbers = dict(
    RED=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9), IR=(10, 11), BG=(12, 13)
)

# Create some compiled regex Patterns to use in this module.
phase_re = re.compile("|".join(phase_names))
orbit_re = re.compile(r"\d{1,6}")
lat_re = re.compile(r"[0-3]?\d?\d?[05]")
target_re = re.compile(fr"{lat_re.pattern}|9(?:[0-2]\d\d|30[0-3])")

obsid_core_re = re.compile(
    fr"(?P<phase>{phase_re.pattern})?_?"
    fr"(?P<orbit>{orbit_re.pattern})_"
    fr"(?P<target>{target_re.pattern})"
)
obsid_re = re.compile(fr"(?<!\w){obsid_core_re.pattern}(?!\d)")

ccd_name_re = re.compile(r"RED|IR|BG")
ccd_re = re.compile(r"RED\d|IR1[0-1]|BG1[2-3]")
ccdid_core_re = re.compile(
    fr"{obsid_core_re.pattern}_(?P<ccd>{ccd_re.pattern})"
)
ccdid_re = re.compile(fr"(?<!\w){ccdid_core_re.pattern}(?!\d)")

chan_re = re.compile(r"(?P<channel>[01])")
ccdchan_re = re.compile(fr"(?P<ccd>{ccd_re.pattern})_(?P<channel>[0-1])")
chanid_core_re = re.compile(fr"{ccdid_core_re.pattern}_(?P<channel>[0-1])")
chanid_re = re.compile(fr"(?<!\w){chanid_core_re.pattern}(?!\d)")

prodid_re = re.compile(
    fr"(?<!\w){obsid_core_re.pattern}_?"
    fr"(?P<ccd>{ccd_re.pattern})?_?"
    fr"(?P<channel>[0-1])?(?!\d)"
)


class ObservationID:
    """A class for HiRISE Observation IDs.

    :ivar phase: The three-letter string indicating the mission phase
        of this Observation ID.
    :ivar orbit_number: The six-digit orbit number (with leading zeros)
        as a string.
    :ivar target: The four digit code (with leading zeroes) is defined as
        follows: Values between 0000 and 3595, inclusive, reflect
        the central latitude of the observation to within a
        half-degree, multiplied by 10.  (0905, for example, is thus
        a latitude of 90.5 degrees.) 0.0 degrees is the night-side
        equator, which is where the orbit number increments. 90.0
        south pole on the ascending pass. 180.0 is the day-side
        equator on the ascending pass. 270.0 is the north pole on
        the descending pass. Text values between 9000 and 9303,
        inclusive, represent off-planet targets such as Phobos,
        Deimos, or stars.
    """

    def __init__(self, *args):
        if len(args) == 1:
            if isinstance(args[0], ObservationID):
                phase = args[0].phase
                orbit = args[0].orbit_number
                target = args[0].target
            else:
                match = obsid_re.search(str(args[0]))
                if match:
                    parsed = match.groupdict()
                    phase = parsed["phase"]
                    orbit = parsed["orbit"]
                    target = parsed["target"]
                else:
                    raise ValueError(
                        f"{args[0]} did not match regex: {obsid_re.pattern}"
                    )
        elif len(args) == 2:
            (phase, orbit, target) = None, *args
        elif len(args) == 3:
            (phase, orbit, target) = args
        else:
            raise IndexError("accepts 1 to 3 arguments")

        self.orbit_number = self.format_orbit(orbit)
        self.target = self.format_target(target)

        if phase is not None:
            if phase in phase_names:
                if is_orbit_in_phase(self.orbit_number, phase):
                    self.phase = phase
                else:
                    raise ValueError(
                        "The orbit, {}, is outside the allowed "
                        "range ({}, {}) in HiRISE mission phase "
                        "{}".format(
                            self.orbit_number,
                            phase_max_orbit[prev_phase(phase)],
                            phase_max_orbit[phase],
                            phase,
                        )
                    )
            else:
                raise ValueError(
                    f"phase argument, {phase}, did not match "
                    f"any known HiRISE phase: {phase_names}"
                )
        else:
            self.phase = get_phase(orbit)

    def __str__(self):
        return f"{self.phase}_{self.orbit_number}_{self.target}"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"'{self.phase}_{self.orbit_number}_{self.target}')"
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.phase == other.phase
                and self.orbit_number == other.orbit_number
                and self.target == other.target
            )
        return False

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return (
                phase_names.index(self.phase),
                int(self.orbit_number),
                int(self.target),
            ) < (
                phase_names.index(other.phase),
                int(other.orbit_number),
                int(other.target),
            )
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self.phase, self.orbit_number, self.target))

    @staticmethod
    def format_orbit(orbit) -> str:
        if orbit_re.fullmatch(str(orbit)):
            return "{:0>6}".format(orbit)
        else:
            raise ValueError(
                f"orbit argument, {orbit}, did not match "
                f"regex: {orbit_re.pattern}"
            )

    @staticmethod
    def format_target(val) -> str:
        if target_re.fullmatch(str(val)):
            return "{:0>4}".format(val)
        else:
            raise ValueError(
                f"target argument, {val}, did not match "
                f"regex: {target_re.pattern}"
            )


class CCDID(ObservationID):
    """A class for HiRISE CCD IDs.

    :ivar ccdname: The CCD name: 'RED', 'IR', or 'BG'.
    :ivar ccdnumber: The CCD number as a string: 0 through 13.
    """

    def __init__(self, *args):
        items = list(args)

        if len(items) == 1:
            if isinstance(args[0], CCDID):
                items = (args[0].phase, args[0].orbit_number, args[0].target)
                ccdname = args[0].ccdname
                ccdnumber = args[0].ccdnumber
            else:
                match = ccdid_re.search(str(items[0]))
                if match:
                    parsed = match.groupdict()
                    items = (
                        parsed["phase"],
                        parsed["orbit"],
                        parsed["target"],
                    )
                    (ccdname, ccdnumber) = get_ccdnamenumber(parsed["ccd"])
                else:
                    raise ValueError(
                        f"Could not construct a CCDID. {args[0]} did not match "
                        f"regex: {ccdid_re.pattern}"
                    )
        elif len(items) == 3:
            try:
                (ccdname, ccdnumber) = get_ccdnamenumber(items.pop())
            except ValueError as err:
                raise ValueError(
                    "Could not construct a CCDID. " + str(err)
                ) from err
        elif len(items) == 4:
            if items[0] in phase_names:
                (ccdname, ccdnumber) = get_ccdnamenumber(items.pop())
            else:
                ccdnumber = _match_num(items.pop())
                ccdname = get_ccdname(items.pop())
        elif len(items) == 5:
            ccdnumber = _match_num(items.pop())
            ccdname = get_ccdname(items.pop())
        else:
            raise IndexError(
                "CCDID accepts 1, 3, 4, or 5 arguments, "
                f"{len(items)} were provided."
            )
        super().__init__(*items)

        self.ccdname = ccdname
        self.ccdnumber = ccdnumber

    def __str__(self):
        return "_".join([super().__str__(), self.get_ccd()])

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.__str__()}')"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                super().__eq__(other)
                and self.ccdname == other.ccdname
                and self.ccdnumber == other.ccdnumber
            )
        return False

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if super().__eq__(other):
                return int(self.ccdnumber) < int(other.ccdnumber)
            else:
                return super().__lt__(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash((super().__hash__(), self.ccdname, self.ccdnumber))

    def get_ccd(self) -> str:
        return self.ccdname + self.ccdnumber

    def get_obsid(self) -> ObservationID:
        return ObservationID(self.phase, self.orbit_number, self.target)


class ChannelID(CCDID):
    """A class for HiRISE Channel IDs.

    :ivar channel: The Channel number (0 or 1) as a string.
    """

    def __init__(self, *args):
        items = list(args)

        if len(items) == 1:
            match = chanid_re.search(str(items[0]))
            if match:
                parsed = match.groupdict()
                items = (parsed["phase"], parsed["orbit"], parsed["target"])
                items += get_ccdnamenumber(parsed["ccd"])
                chan = parsed["channel"]
            else:
                raise ValueError(
                    "{} did not match regex: "
                    "{}".format(args[0], chanid_re.pattern)
                )
        elif 2 <= len(items) <= 6:
            maybechan = items.pop()
            try:
                chan = self.format_chan(maybechan)
            except ValueError as err:
                raise ValueError("The last item, " + str(err))

            if len(items) == 1:
                if isinstance(items[0], CCDID):
                    items = [
                        items[0].phase,
                        items[0].orbit_number,
                        items[0].target,
                        items[0].ccdname,
                        items[0].ccdnumber,
                    ]
                else:
                    raise ValueError(
                        "{} was not a CCDID object".format(items[0])
                    )

            elif len(items) == 2:
                (ccdname, ccdnumber) = get_ccdnamenumber(items.pop())
                if isinstance(items[0], ObservationID):
                    items = [
                        items[0].phase,
                        items[0].orbit_number,
                        items[0].target,
                        ccdname,
                        ccdnumber,
                    ]
                else:
                    raise ValueError(
                        "{} was not a ObservationID object".format(items[0])
                    )

        else:
            raise IndexError("accepts 1 to 6 arguments")
        super().__init__(*items)

        self.channel = str(chan)

    def __str__(self):
        return "_".join([super().__str__(), self.channel])

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.__str__()}')"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return super().__eq__(other) and self.channel == other.channel
        return False

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if super().__eq__(other):
                return int(self.channel) < int(other.channel)
            else:
                return super().__lt__(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash((super().__hash__(), self.channel))

    @staticmethod
    def format_chan(chan) -> str:
        if isinstance(chan, int):
            if chan == 0 or chan == 1:
                return chan
            else:
                raise ValueError(f"{chan}, is not zero or one.")
        else:
            matched = chan_re.fullmatch(chan)
            if matched:
                return matched.group()
            else:
                raise ValueError(
                    f"{chan}, did not match a " "channel {chan_re.pattern}"
                )


# def _match_chan(s: str):
#     matched = chan_re.fullmatch(s)
#     if matched:
#         return matched.group()
#     else:
#         return None


def _match_num(s: str):
    matched = re.fullmatch(r"1?\d", s)
    if matched:
        return matched.group()
    else:
        return None


# def _namenumchan(one: str, two: str) -> tuple:
#     ccdname = None
#     ccdnumber = None
#     chan = None
#     try:
#         (ccdname, ccdnumber) = get_ccdnamenumber(one)
#         chan = _match_chan(two)
#     except ValueError:
#         ccdname = get_ccdname(one)
#         ccdnumber = _match_num(two)
#     return (ccdname, ccdnumber, chan)


def parseObsID(s: str) -> tuple:
    """Parses a string to find the first occurence of a valid ObsID.

    Returns a tuple of strings that were matched.
    """
    match = obsid_re.search(str(s))
    if match:
        return match.groups()
    else:
        raise ValueError(
            "{} did not match regex: {}".format(s, obsid_re.pattern)
        )


def prev_phase(s: str) -> str:
    """Returns the name of the previous HiRISE mission phase."""
    prev_index = phase_names.index(s) - 1
    if prev_index < 0:
        raise IndexError(
            f"HiRISE mission phase, {s}, is the first one, "
            "it has no previous phase."
        )
    else:
        return phase_names[prev_index]


def is_orbit_in_phase(orbit, phase: str):
    """Determines whether the given orbit is within the given HiRISE
    mission phase."""
    previous = prev_phase(phase)
    if int(orbit) <= phase_max_orbit[previous]:
        if (
            phase_max_orbit[phase] is not None
            and phase_max_orbit[phase] == phase_max_orbit[previous]
            and phase_max_orbit[phase] == int(orbit)
        ):
            return True
        return False
    if (
        phase_max_orbit[phase] is not None
        and int(orbit) > phase_max_orbit[phase]
    ):
        return False
    return True


def ObsIDstr(s: str) -> str:
    """Extracts a HiRISE Observation ID string from the given string."""
    return str(ObservationID(s))


def get_phase(orbit) -> str:
    """Returns the name of the HiRISE mission phase based on the MRO orbit
    number.

    The given orbit must be an int or something that can be converted
    to an int via int().
    """
    last_phase = None
    for (k, v) in reversed(list(phase_max_orbit.items())):
        if v is not None and int(orbit) > v:
            if last_phase is not None:
                return last_phase
            else:
                raise IndexError(
                    f"The orbit number {orbit} is greater than "
                    "the last HiRISE orbit."
                )
        last_phase = k
    else:
        raise IndexError(
            f"The orbit number {orbit} is not a HiRISE orbit " "number."
        )


def get_ccd(s: str) -> str:
    """Extracts a HiRISE CCD from the string (would get
    'RED5' from 'PSP_010502_2090_RED5_0.EDR_Stats.cub')."""
    match = ccd_re.search(s)
    if match:
        return match.group()
    else:
        raise ValueError(f"{s} did not match regex: {ccd_re.pattern}")


def _getccdname_fromint(ccdnum: int) -> str:
    for k, v in ccd_numbers.items():
        if ccdnum in v:
            return k
    else:
        raise ValueError(f"The value {ccdnum} must be a value from 0 to 13.")


def get_ccdname(item) -> str:
    """Extracts the name of a HiRISE CCD from the string (would get
    'RED' from 'PSP_010502_2090_RED5_0.EDR_Stats.cub').  If an integer
    is given instead of a string, it will provide the corresponding name.
    """
    try:
        match = ccd_name_re.search(item)
        if match:
            return match.group()
        else:
            try:
                return _getccdname_fromint(int(item))
            except ValueError:
                raise ValueError(
                    f"{item} did not match regex: {ccd_name_re.pattern}"
                )
    except TypeError:
        try:
            if isinstance(item, int):
                return _getccdname_fromint(int(item))
            else:
                raise TypeError("Expected string or integer object.")
        except ValueError:
            raise


def get_ccdnumber(s: str) -> str:
    """Extracts the number of a HiRISE CCD from the string (would get
    '5' from 'PSP_010502_2090_RED5_0.EDR_Stats.cub')."""
    match = ccd_re.search(s)
    if match:
        return re.search(r"\d{1,2}", match.group()).group()
    else:
        raise ValueError(f"{s} did not match regex: {ccd_re.pattern}")


def get_ccdnamenumber(item) -> tuple:
    """Extracts the name and number of a HiRISE CCD from the input (would get
    '('RED', '5')' from 'PSP_010502_2090_RED5_0.EDR_Stats.cub')."""
    try:
        nn = get_ccd(item)
        return get_ccdname(nn), get_ccdnumber(nn)
    except TypeError:
        return get_ccdname(item), str(item)
    except ValueError:
        return get_ccdname(int(item)), item


def get_ccdchannel(s: str) -> tuple:
    """Extracts a HiRISE CCD and channel number from the string (would get
    ('RED5', '1') from 'PSP_010502_2090_RED5_0.EDR_Stats.cub')."""
    match = ccdchan_re.search(s)
    if match:
        return match.groups()
    else:
        raise ValueError(f"{s} did not match regex: {ccdchan_re.pattern}")


def _reverse_recurse(path: os.PathLike, IDclass, name):
    p = Path(path)
    for part in reversed(p.parts):
        try:
            return IDclass(part)
        except ValueError:
            continue
    raise ValueError(f"Could not extract a {name} from {path}")


def _get_fromfile(path: os.PathLike, IDclass, name, archivekey):
    p = Path(path)

    try:
        return IDclass(isis.getkey_k(p, "Archive", archivekey))
    except (subprocess.CalledProcessError, ValueError):
        # The CalledProcessError is if there is some problem with running
        # getkey, the ValueError is if a CCDID can't be extracted from
        # the labels.  This allows this function to be called on any kind of
        # file, as it will just try and read the CCDID from the filename
        # or could also reverse recurse up to directory names.
        return _reverse_recurse(p, IDclass, name)


def get_ObsID_fromfile(path: os.PathLike) -> ObservationID:
    """Reads the file to get the ObservationID, if an ISIS cube,
       otherwise parses the filepath."""

    return _get_fromfile(
        path, ObservationID, "HiRISE Observation ID", "ObservationId"
    )


def get_CCDID_fromfile(path: os.PathLike) -> CCDID:
    """Reads the file to get the CCDID, if an ISIS cube,
       otherwise parses the filepath."""

    return _get_fromfile(path, CCDID, "HiRISE CCD ID", "ProductId")


def get_ChannelID_fromfile(path: os.PathLike) -> ChannelID:
    """Reads the file to get the ChannelID, if an ISIS cube,
       otherwise parses the filepath."""

    return _get_fromfile(path, ChannelID, "HiRISE Channel ID", "ProductId")
